#!/usr/bin/perl
# Filename:	signup.cgi
# Author:	David Ljung Madison <DaveSource.com>
# See License:	http://MarginalHacks.com/License
# Version:	1.00
# Description:	Signup board for volunteers
#		Good for a simple board where users are essentially trusted.
# Installation:
#
# Make sure CGI scripts are turned on in your httpd configuration
# Put signup.cgi into your CGI directory.
# Make sure it's executable by httpd user (or just: "chmod ugo+rx signup.cgi")
# Put signup_data somewhere readable by httpd user (such as cgi-bin)
# Make sure it's writable by httpd user
# (or just: "chmod ugo+rwx signup_data ; chmod -R ugo+rw signup_data")
# Change "SETTINGS" below as required
#
# Format:	The database format is tab-separated-values:
#
#   database_line := heading | shift
#   heading := <heading> | <heading> TAB <heading2>
#   shift   := <time> TAB <job> TAB <description> TAB <num needed> volunteers*
#   volunteers := TAB <name> TAB <email>
#
# time, job, description are all just strings (can be empty)
# num needed is the number of signups available for a given shift
# 
# You should also create a Mail_Info file.  See MAIL_INFO below.
#
use strict;
use POSIX;	# For mktime

umask 0022;

##################################################
##################################################
#
# SETTINGS (you probably want to change these)
#
##################################################
##################################################

## The name of your machine
## Change the default at the end:
my $HOST = $ENV{HTTP_HOST} || $ENV{SERVER_NAME} || 'YOURDOMAINHERE.com';

## Port (from ENV or from HOST or 80)
my $PORT = $ENV{SERVER_PORT} || $ENV{HTTP_PORT};
   $PORT = (($HOST =~ s/:(\d+)$//) ? $1 : 80) unless (defined $PORT);

## For the URL
my $DOMAIN = $PORT==80 ? $HOST : "$HOST:$PORT";

## If you're using SSL, here's where you choose 'https:'
my $HTTP = 'http';
#my $HTTP = 'https';

## The location of the cgi script (we might be guessing it if
## the REFERER not specified)
my $CGI = $ENV{REQUEST_URI} || '/cgi-bin/signup.cgi';
   $CGI =~ s/[\&\?][^\/]+$//;

## The data directory (Path, not URL)
my $DATA =	'/usr/lib/cgi-bin/signup_data/';

## Mail info contains information for constructing outgoing mail.
## It has three lines:  Event name, Mail from full name, Mail from address
my $MAIL_INFO = "$DATA/Mail_Info";

## Volunteer database
my $VOLUNTEERS =	"$DATA/Volunteers.tsv";	# Put the data here..

my $HOURLY_SAVE =	1;	# Make saves (once/hour) of database when it changes

## Where is sendmail? (forge_pipe is best!)
my $SENDMAIL  = '/usr/bin/forge_pipe';  # See http://MarginalHacks.com/#forge
   $SENDMAIL  = '/usr/sbin/forge_pipe' if (! -x $SENDMAIL);
   $SENDMAIL  = '/usr/sbin/sendmail' if (! -x $SENDMAIL);
   $SENDMAIL  = '/usr/lib/sendmail' if (! -x $SENDMAIL);
   $SENDMAIL  = '/usr/bin/sendmail' if (! -x $SENDMAIL);

my $EVENT = 'SFLX Volunteer';
my $MAIL_FROM = 'luislew@hotmail.com';
my $MAIL_FROM_FULL = 'SFLX Volunteers Mail';

## When do cookies expire?
		my $HOUR = 60*60;
		my $DAY = 24*$HOUR;
		my $WEEK = 7*$DAY;
		my $MONTH = 31*$DAY;    # kludgy
		my $YEAR = 365*$DAY;

my $EXPIRES = 3*$MONTH;		# Three months from now

# Show how many posts at once?
my $MAX_POSTS = 10;

# Where are the signup images (URL, not path)
my $IMAGES = '/images';

##################################################
##################################################
#
# Code
#
##################################################
##################################################
# These you probably shouldn't change
##################################################
my $URL = "${HTTP}://$DOMAIN$CGI";

# Use javascript to hide mailto links so they can't be gathered by spammers
my $OBSCURE_MAILTO = 0;

##################################################
# Query
##################################################
sub from_url {
  my ($str) = @_;
  $str =~ s/\+/ /g;
  $str =~ s/%([0-9a-f]{2})/chr(hex($1))/eig;
  $str;
}

sub to_url {
  my ($str) = @_;
  $str =~ s/([^ a-zA-Z0-9\.])/"%".sprintf("%0.2x",ord($1))/eg;
  $str =~ s/ /+/g;
  $str;
}

sub unhtml {
  my ($str) = @_;
  $str = from_url($str);
  $str =~ s/</\&lt;/g;
  $str =~ s/>/\&gt;/g;
  $str =~ s/[\r\n]/ /g;
  $str;
}

sub html {
  my (@str) = @_;
  my $str = join("<br>\n",split("\n",join("",@str)));
  $str;
}

my %COOKIE;
sub parse_query {
  # Get query
  my $query_string;
  if ($ENV{REQUEST_METHOD} eq "POST") {
    read(STDIN,$query_string,$ENV{CONTENT_LENGTH});
  } elsif ($ENV{QUERY_STRING}) {
    $query_string = $ENV{QUERY_STRING};
  } elsif (@ARGV) {
    $query_string = join("&",@ARGV);
  }
  chomp($query_string);

  # Split query
  # $query_string is of the form:  "variable=value&var2=val2&.."
  my @querys=split(/[\&\?]/,$query_string);
  my (%query,$var,$val);
  foreach my $str (@querys) {
    $var = $str if (!(($var,$val) = ($str =~ /([^=]*)=(.*)/)));
    $val = 1 unless defined $val;
    $query{$var} = from_url($val);
  }

  # Handle cookies
  foreach my $str ( split("; ?",$ENV{HTTP_COOKIE}) ) {
    $str =~ s/\s*(.*?)\s*/$1/;
    $var = $str if (!(($var,$val) = ($str =~ /([^=]*)=(.*)/)));
    $val = 1 unless defined $val;
    $COOKIE{$var} = from_url($val);
  }

#header(); show_values(\%query);
  \%query;
}

sub show_values {
  my ($query) = @_;
  print "<p><hr><p>\n\n";
  foreach my $q ( keys %$query ) {
    print "yo $q -> $query->{$q}<br>\n";
  }
}

##################################################
# Cookies
##################################################
sub send_cookies {
  my ($name,$email) = @_;
  my ($uname,$uemail) = (to_url($name),to_url($email));
  my $expires = strftime "%a, %d-%b-%Y %H:%M:%S %Z", localtime(time+ $EXPIRES);
  print "Set-Cookie: name=$uname; $expires\n" if ($name ne $COOKIE{name});
  print "Set-Cookie: email=$uemail; $expires\n" if ($email ne $COOKIE{email});
  $COOKIE{name} = $name;
  $COOKIE{email} = $email;
}

##################################################
# HTML
##################################################
sub cat {
  my ($file) = @_;

  open(FILE,"<$file") || return;
  while (<FILE>) { print; }
  close FILE;
}

my $DID_HEADER=0;
sub header {
  my ($query) = @_;

  # Set cookies on preview or post
  send_cookies($query->{name},$query->{email})
    if $query->{signup} && !bad_name_email($query);

  return if $DID_HEADER++;
  print "Content-type: text/html\n\n";

  cat("$DATA/Header.html");
}

#sub ERROR { header(); print "<h2>ERROR: @_</h2>\n"; footer(); undef; }
sub ERROR { header(); print "<h2><font color=red>ERROR: @_</font></h2>\n"; undef; }

sub footer {
  return unless $DID_HEADER;
  cat("$DATA/Footer.html");
}

##################################################
# Mail
##################################################
# Send mail
sub send_mail {
  my ($to,$cc,$subject,$from,$from_full,$msg) = @_;

  return ERROR("Can't open sendmail pipe!")
    unless open(MAIL,"|$SENDMAIL -i -f $from -F \"$from_full\" -t");

  $to.="\nCc: $cc" if $cc && $cc ne $to;

  print MAIL <<MAIL_MESSAGE;
From: "$from_full" <$from>
Subject: $subject
To: $to
X-Mailer: $URL

$msg

MAIL_MESSAGE
  close MAIL;

  return 1;
}

##################################################
# Database
##################################################
sub read_volunteers {
  my (@vol);
  open(VOL,"<$VOLUNTEERS") || return ERROR("Can't read database [$VOLUNTEERS]");
  while (<VOL>) {
    chomp;
    next unless /\S/;
    my %row;
    push(@vol,\%row);
    my @row = split(/\t/);
    @{$row{cols}} = @row;
    if ($#row<2) {
      $row{type} = 'heading';
      next;
    }
    $row{type} = 'shift';
    ($row{time},$row{job},$row{desc},$row{num}) =
      (shift @row, shift @row, shift @row, shift @row);
    # Rest is volunteers, name \t email
    while (@row) {
      push(@{$row{taken}}, [shift @row, shift @row]);
    }
  }
  close VOL;
  \@vol;
}

sub my_sys { my ($cmd) = @_;  system($cmd);  $?==0?1:0; }

sub write_volunteers {
  my ($query,$vol,$attempt) = @_;

  # (For blocking open)
  sleep 1 if ($attempt>4);
  return ERROR("Couldn't write new volunteer database - try again?")
    if ($attempt>6);

  my $new = "$VOLUNTEERS.new";

  # Blocking open
  sysopen(N, $new, O_WRONLY | O_EXCL | O_CREAT) 
    || return write_volunteers($query,$vol,$attempt+1);

  foreach my $row ( @$vol ) {
    my @cols = @{$row->{cols}};
    print N join("\t",@cols),"\n";
  }
  close N;

  # Save a copy?
  if ($HOURLY_SAVE) {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my $time = sprintf("%d.%2.2d.%2.2d.%2.2d.%2.2d",$year+1900,$mon+1,$mday,$hour,$min);
    my $copy = "$VOLUNTEERS.$time";
    system("/bin/cp $VOLUNTEERS $copy");
  }

  # This could be serious because it could leave behind an old ".new" file
  # which will block new writes...
  if (!rename($new,$VOLUNTEERS) && !my_sys("/bin/mv $new $VOLUNTEERS")) {
    ERROR("Couldn't overwrite volunteer db! [$VOLUNTEERS]");
    unlink($new) || ERROR("And couldn't remove new volunteer db!  Trouble!");
    return undef;
  }
  return 1;
}

##################################################
# Code
##################################################
my $DID_OBSCURE_FUNCTION = 0;
sub obscure_email {
  my ($user,$text) = @_;
  my $ret = "";

  my $domain = "";
  ($user,$domain) = ($1,$2) if $user =~ /(.+)\@(.+)/;

  unless ($DID_OBSCURE_FUNCTION++) {
    $ret .= <<OBSCURE_FUNCTION;
<script type='text/javascript' language='JavaScript'>
<!--
  function  emilN( user, dom, tn ) {
     output = '<a ID="mail" class="footnote" href="mai' + 'lto:' + user + '\@' + dom + '">' + tn + '</a>';
     document.write(output);
  }
//-->
</script>
OBSCURE_FUNCTION
  }

  $ret .= <<OBSCURE_CALL;
<script type='text/javascript' language='JavaScript'>
<!--
emilN('$user', '$domain', '$text');
//-->
</script>
OBSCURE_CALL

  $ret;
}

sub signup_form {
  my ($query,$num) = @_;

  print <<FORM;
      <div style="white-space: nowrap">
        <form action=$URL method=POST name=signmeup$num>
          <input type=hidden name=signup value='$num'>
FORM

  if ($COOKIE{email}) {
          #<i><font color='#aaaaaa'>$COOKIE{name}</font></i>
    print <<FORM_COOKIE;
          <input type=hidden name=name value='$COOKIE{name}'>
          <input type=hidden name=email value='$COOKIE{email}'>
FORM_COOKIE
  } else {
    print <<FORM_NO_COOKIE;
          <input type=text name=name size=8 maxsize=50 value='name'>
          <input type=text name=email size=12 maxsize=50 value='email'>
FORM_NO_COOKIE
  }

  print <<END_FORM;
    <a href="javascript:document.signmeup$num.submit()"
         onMouseOver = document.images['SignMeUp$num'].src='$IMAGES/SignMeUp.down.gif'
         onMouseOut = document.images['SignMeUp$num'].src='$IMAGES/SignMeUp.gif'
         onMouseDown = document.images['SignMeUp$num'].src='$IMAGES/SignMeUp.down.gif'
         onMouseUp = document.images['SignMeUp$num'].src='$IMAGES/SignMeUp.gif'>
      <img src='$IMAGES/SignMeUp.gif' alt="Sign Me Up" width=69 height=19 name=SignMeUp$num border=0></a>
  </form>
</div>
END_FORM
}

sub bad_name_email {
  my ($query) = @_;

  $query->{email} =~ s/^\s*//;
  $query->{email} =~ s/\s*$//;

  return "Please enter your name before clicking 'Sign Me Up'"
    if $query->{name} eq "name";
  return "Please enter your email before clicking 'Sign Me Up'"
    if $query->{email} eq "email";
  return "Email address can't have spaces [$query->{email}]"
    if $query->{email} =~ /\s/;
  return "Email must be of the form 'name\@domain.com' [$query->{email}]"
    if $query->{email} !~ /^\s*\S+\@\S+\.\S+\s*$/;
  return 0;
}

sub signup {
  my ($query,$vol) = @_;

  return unless $query->{signup};
  my ($num,$name,$email) = ($query->{signup},$query->{name},$query->{email});

  my $err = bad_name_email($query);
  return ERROR($err) if $err;

  return ERROR("Bad signup value?  Should be a number [$num]")
    unless $num =~ /^\d+$/;
  my $row = $vol->[$num];
  return ERROR("Bad signup value?  Not a volunteer column? [$num]")
    unless $row->{type} eq 'shift';
  my $got = $row->{taken} ? ($#{$row->{taken}}+1) : 0;

  # Are they already signed up?
  foreach my $volunteer ( @{$row->{taken}} ) {
    return ERROR("Already signed up for job [$row->{job}]")
      if $email eq $volunteer->[1];
  }

  return ERROR("Job [$row->{job}] already has enough volunteers")
    unless $row->{num} > $got;

  push(@{$row->{taken}}, [$name, $email]);
  push(@{$row->{cols}}, $name, $email);

  unless (write_volunteers($query,$vol)) {
    pop(@{$row->{taken}});
    pop(@{$row->{cols}});
    pop(@{$row->{cols}});
    return;
  }
  signup_mail($name,$email,$row->{job},$row->{description});
}

sub signup_mail {
  my ($name,$email,$job,$description) = @_;

  my $who = $name || $email;

  my $underline = "-" x (length($EVENT));

  # Get the mail info
  my ($event,$from,$from_full) = ("Volunteer","Volunteer mail",$email);
  if (open(M,"<$MAIL_INFO")) {
    $event = scalar <M>;
    $from_full = scalar <M>;
    $from = scalar <M>;
    close M;
    chomp($event); chomp($from); chomp($from_full);
  }

  # Send it.
  send_mail($email,0,"$EVENT Signup", $from, $from_full,<<MSG);

Thank you $who!

This email ($email) was signed up for:

$EVENT
$underline
Job:  $job

$description

If this was in error or this needs to be changed, please let us know!

Otherwise, thanks for signing up!

----------------------------------------------------------------------
Delivered by the $EVENT Signup Sheet.
Location:   $URL
Script by:  http://MarginalHacks.com/
----------------------------------------------------------------------
MSG
}

sub show_volunteers {
  my ($query,$vol) = @_;

  print "<table width='800'>\n";
  print "  <tr valign=top>\n";
  print "    <td><b>Time</b></td>\n";
  print "    <td width='10%'><b>Title</b></td>\n";
  print "    <td><b>Description</b></td>\n";
  print "    <td align=right><b>#</b>&nbsp;</td>\n";
  print "    <td><b>Volunteers</b></td>\n";
  print "  </tr>\n";
  print "  <tr>\n";
  print "    <td colspan=5><hr></td>\n";
  print "  </tr>\n";

  for (my $v=0; $v<=$#$vol; $v++) {
    my $row = $vol->[$v];
    if ($row->{type} eq 'heading') {
      print "  <tr>\n";
      foreach my $col ( @{$row->{cols}} ) {
        print "    <td bgcolor=yellow><b>$col</b></td>\n";
      }
      print "    <td> </td>\n" x (4-$#{$row->{cols}});
      print "  </tr>\n";
      next;
    }
    print "  <tr valign=top>\n";
    print "    <td>$row->{time}</td>\n";
    print "    <td>$row->{job}</td>\n";
    print "    <td><font size='-1'>$row->{desc}</font></td>\n";
    print "    <td width='5%' align=right><b>$row->{num}</b>&nbsp;</td>\n";
    print "    <td>\n";
    my $already_signed_up = 0;
    foreach my $volunteer ( @{$row->{taken}} ) {
      print "      ".obscure_email($volunteer->[1],$volunteer->[0])."<br>\n";
      $already_signed_up=1 if $COOKIE{email} && $COOKIE{email} eq $volunteer->[1];
    }
    signup_form($query,$v)
      if ($#{$row->{taken}}+1 < $row->{num} && !$already_signed_up);
    print "    </td>\n";
    print "  </tr>\n";
  }

  print "</table>\n";
}

##################################################
# Do it
##################################################
sub main {
  my $query = parse_query();

  header($query);

  my $vol = read_volunteers();
  signup($query,$vol);
  show_volunteers($query,$vol);

  footer();

} main();
